; sonapic.asm
; PIC16C54 based sonar ranging interface for Polaroid module
;
; Copyright 1994, 1995 Eric Smith
;
; sonapic.asm is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License version 2 as published by the
; Free Software Foundation.  Note that I am not granting permission to
; redistribute or modify sonapic.asm under the terms of any later version of
; the General Public License.
;
; This program is distributed in the hope that it will be useful (or at least
; amusing), but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
; Public License for more details.
;
; You should have received a copy of the GNU General Public License along with
; this program (in the file "COPYING"); if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; $Header: /usr/home/kolwynia/eric/pic/sonapic/RCS/sonapic.asm,v 1.2 1995/09/16 12:49:27 eric Exp $


	DEVICE	PIC16C54,HS_OSC,WDT_OFF,PROTECT_OFF


f	equ	1		;for destination argument
w	equ	0		;for destination argument

ind	equ	0		;used for indirects thru fsr
rtcc	equ	1		;real time clock/counter
pcl	equ	2		;program counter
fsr	equ	4		;file select register (index register)

status	equ	3		;status register:
cf	equ	0		;  bit 0 = carry bit flag
dcf	equ	1		;  bit 1 = digit carry bit flag
zf	equ	2		;  bit 2 = zero bit flag
pdf	equ	3		;  bit 3 = power down bit flag
tof	equ	4		;  bit 4 = time out bit flag
				;  bits 5,6,7 do not apply to pic16c54


nothing	equ	0		;used to return nothing from subroutines


porta	equ	5		;i/o port ra:
				;  bit 0 unused
rxd	equ	1		;  bit 1 = serial receive
txd	equ	2		;  bit 2 = serial transmit
pzt	equ	3		;  bit 3 = pzt speaker

portb	equ	6		;i/o port rb:
echo	equ	0		;  bit 0 = echo from polaroid module
init	equ	1		;  bit 1 = init to polaroid module
binh	equ	2		;  bit 2 = binh (???), drive low
blnk	equ	3		;  bit 3 = blnk (???), drive low
				;  bits 4-7 unused


ram	equ	07h		;start of ram
body	equ	000h		;beginning of program


	org	ram		;start of ram

one	ds	1		;holdes the value of "1" for incrementing w-reg
zero	ds	1		; holds the value 0 for doing 2s complement

count	ds	1		;general purpose counter
dlycnt	ds	1		;counter for delays (used in several places)
mscnt	ds	1		;number of milliseconds using delay routine
cycles	ds	1		; number of cycles in a beep
period	ds	1		; period of a click
temp	ds	1		; very temporary

range:	ds	2
maxrange	equ	10668	; maximum range in mm (35 ft)

serreg:	ds	1		; character being transmitted
bitcnt:	ds	1

speed	equ	104-3		; 4800 baud (delay count for one bit time)

	if	1
count2	ds	1
r0	ds	1
r1	ds	1
r2	ds	1
	else
count2	equ	dlycnt
r0	equ	mscnt
r1	equ	cycles
r2	equ	period
	endif

acca	equ	range
accb	equ	dlycnt
accc	equ	cycles
accd	equ	serreg


conv	equ	25802		; conversion from mm to tenths of inches
				; (65536 / 2.54)


	org	body		;start of rom


;call extenders

reset:	goto	resetx
delay:	goto	delayx
beep:	goto	beepx
click:	goto	clickx


hextbl:	andlw	0fh
	addwf	pcl
	retw	'0','1','2','3','4','5','6','7'
	retw	'8','9','a','b','c','d','e','f'


msgtbl:	addwf	pcl
rngmsg	equ	(.-msgtbl)-1
	retw	'O','u','t',' ','o','f',' ','r','a','n','g','e',0
mmsg	equ	(.-msgtbl)-1
	retw	' ','m',0dh,0ah,0
inmsg	equ	(.-msgtbl)-1
	retw	' ','i','n',0dh,0ah,0
hexmsg1	equ	(.-msgtbl)-1
	retw	'(','$',0
hexmsg2	equ	(.-msgtbl)-1
	retw	')',' ',0


resetx:
;	clrwdt			;reset watchdog timer
;	movlw	3fh		;external edge to timer
;	option			;high to low edge for timer
				;prescaler assigned to watchdog
				;prescaler divide by 128 for now

	clrf	fsr		; just to be safe

	clrf	zero		; initialize the value of zero

	movlw	1		;initialize the value of one to allowing
	movwf	one		;incrementing the w register using "addwf one,w"

	movlw	(1<<rxd)	;porta:
	tris	porta		;  bit 0 = unused (output)
				;  bit 1 = rxd (input)
				;  bit 2 = txd (output)
				;  bit 3 = pzt (output)
				;  bits 4-7 = inputs (no pins on 16c54)

	bcf	porta,pzt	;0 volts on speaker

	movlw	000h		;portb:
	movwf	portb		;drive port b output pins low
	movlw	001h		; except b0 (echo input)
	tris	portb

	movlw	00ch		; clear display
	call	xmit

	call	x30

	call	beep

main:
	clrf	range
	clrf	range+1
	bsf	portb,init

; delay 1 millimeter = 6 microseconds = 12 cycles at 8.0 MHz
rdloop:
	incf	range		; 0 increment range
	btfsc	status,zf	; 1
	incf	range+1		; 2
	movf	range+1,w	; 3 test for timeout - 10752 mm
	xorlw	(maxrange/256)+1	; 4
	btfsc	status,zf	; 5
	goto	rddone		; 6
	btfss	portb,echo	; 7 test for echo
	goto	rdloop		; 8
rddone:

	bcf	portb,init


	movlw	00ch		; clear display
	call	xmit

	call	x30


; display count in hexadecimal
	movlw	hexmsg1
	call	outmsg
	movf	range+1,w
	call	hout2
	movf	range,w
	call	hout2
	movlw	hexmsg2
	call	outmsg

	movf	range+1,w
	xorlw	(maxrange/256)+1
	btfsc	status,zf
	goto	outofrange

; display count in meters
	movf	range,w
	movwf	accb
	movf	range+1,w
	movwf	accb+1
	call	bin2bcd

	movf	r2,w
	call	hextbl
	call	xmit

	swapf	r1,w
	call	hextbl
	call	xmit

	movlw	'.'
	call	xmit

	movf	r1,w
	call	hextbl
	call	xmit

	movf	r0,w
	call	hout2

	movlw	mmsg
	call	outmsg


; display count in inches
	movlw	(conv&0ffh)
	movwf	accb
	movlw	(conv/256)
	movwf	accb+1
	call	dmul
	call	bin2bcd

	movf	r1,w
	call	hout2

	swapf	r0,w
	call	hextbl
	call	xmit

	movlw	'.'
	call	xmit

	movf	r0,w
	call	hextbl
	call	xmit

	movlw	inmsg
	call	outmsg

	goto	wait


outofrange:
	movlw	rngmsg
	call	outmsg

wait:
	movlw	200		; delay for 1/5 second (8 MHz)
	call	delay
	movlw	200
	call	delay

	movlw	200		; delay for 1/5 second (8 MHz)
	call	delay
	movlw	200
	call	delay

	movlw	200		; delay for 1/5 second (8 MHz)
	call	delay
	movlw	200
	call	delay

	movlw	200		; delay for 1/5 second (8 MHz)
	call	delay
	movlw	200
	call	delay

	movlw	200		; delay for 1/5 second (8 MHz)
	call	delay
	movlw	200
	call	delay

	goto	main


outmsg:
	movwf	temp
outmsg1:
	movf	temp,w
	call	msgtbl
	iorlw	0		; set zero flag as appropriate
	btfsc	status,zf
	retlw	nothing
	call	xmit
	incf	temp
	goto	outmsg1


hout2:	movwf	temp
	swapf	temp,w
	call	hextbl
	call	xmit
	movf	temp,w
	call	hextbl
; fall into xmit


; Transmit character in w-reg as 8 bits, no parity, 1 stop.
; This routine has the start, stop and data bits inverted because an inverting
; RS-232 buffer is not used.
;
xmit:	movwf	serreg		;save character
	movlw	10		;put # of data bits + start bit + # stop bits
	movwf	bitcnt		; into counter
	bsf	status,cf	;set up the stop bit
	bsf	porta,txd	;send start bit

xmt10:	movlw	speed		;delay 1 bit time
	movwf	dlycnt
xd1:	nop			;				1 cycle
	decfsz	dlycnt		;				1
	goto	xd1		;				2 =4 cycles

	decf	bitcnt		;count the data bits
	btfsc	status,zf
	goto	xmt30		; and exit when done
	rrf	serreg		;get data bit into carry
	btfsc	status,cf	;if carry is set
	bcf	porta,txd	; then xmit a zero
	btfss	status,cf	;if carry is clear
	bsf	porta,txd	; then transmit a one
	goto	xmt10

xmt30:
x30:	movlw	10		; extra delay to try to make LCD happy
	movwf	bitcnt
x31:	movlw	speed
	movwf	dlycnt
x32:	nop
	decfsz	dlycnt
	goto	x32
	decfsz	bitcnt
	goto	x31
	retlw	nothing		;end of "xmit"


; convert binary in accb+1:accb to bcd in r2:r1:r0
; from Microchip ap note #526, "Utility Math Routines", document number
; DS00526C, page 51
; NOTE: changed order of result, now r2 is most significant
bin2bcd:
	bcf	status,cf
	movlw	16
	movwf	count
	clrf	r0
	clrf	r1
	clrf	r2

loop16:	rlf	accb
	rlf	accb+1
	rlf	r0
	rlf	r1
	rlf	r2

	decfsz	count
	goto	adjdec
	retlw	nothing


adjdec:	movlw	3
	movwf	count2

	movlw	r0
	movwf	fsr

adjbcd:	movlw	3
	addwf	ind,w
	movwf	temp
	btfsc	temp,3
	movwf	ind
	movlw	30h
	addwf	ind,w
	movwf	temp
	btfsc	temp,7
	movwf	ind
	incf	fsr
	decfsz	count2
	goto	adjbcd
	goto	loop16



; double precision unsigned multiple of acca+1:acca by accb+1:accb
; with result in accb+1:accb:accc+1:accc
; from Microchip ap note #526, "Utility Math Routines", document number
; DS00526C, page 11

dmul:
	movlw	16		; temp = 16
	movwf	temp

	movf	accb+1,w	; accd = accb
	movwf	accd+1
	movf	accb,w
	movwf	accd

	clrf	accb+1		; accb = 0
	clrf	accb

mloop:	rrf	accd+1
	rrf	accd
	btfss	status,cf	; need to add?
	goto	noadd		; nope

	movf	acca,w		; accb += acca (+1?)
	addwf	accb

	btfsc	status,cf
	incf	accb+1

	movf	acca+1,w
	addwf	accb+1

noadd:	rrf	accb+1		; does carry in matter???
	rrf	accb
	rrf	accc+1
	rrf	accc

	decfsz	temp
	goto	mloop

	retlw	nothing


;delay for w *1ms at 4mhz osc. freq.

delayx:	movwf	mscnt		;count milliseconds
dly10:	movlw	250
	movwf	dlycnt
dly20:	decf	dlycnt		;				1 cycle
	btfss	status,zf	;				1
	goto	dly20		;				2 =4 cycles
	decfsz	mscnt		;last msec?
	goto	dly10
	retlw	nothing


;make a "bell" sound

beepx:	movlw	200
	movwf	cycles
bp10:	movlw	254		;a nice tone???
	movwf	period		;(fall into "click")

;make a "click" sound "cycles" number of times.
;	frequency = 1 /(1 /(osc frequency /4) *5 cycles *period *2)

clickx:	bsf	porta,pzt	;click high
	movf	period,w	;time for one half of cycle
	movwf	dlycnt		; into delay counter
clk10:	goto	clk15		;(burn 2 cycles)		2 cycles
clk15:	decfsz	dlycnt		;				1
	goto	clk10		;				2= 5 cycles
	bcf	porta,pzt	;click low
	movf	period,w	;time for one half of cycle
	movwf	dlycnt		; into delay counter
clk20:	goto	clk25		;(burn 2 cycles)		2 cycles
clk25:	decfsz	dlycnt		;				1
	goto	clk20		;				2= 5 cycles

	decfsz	cycles		;another cycle of click?
	goto	click		; branch if so
	retlw	nothing

